/***************************************************************************
                          tea_encode.c  -  description
                             -------------------
    begin                : 
    copyright            : (C) 2006 by Peter 'Roxton' Semiletov
    email                : peter.semiletov@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include "tea_encode.h"
#include "tea_defs.h"
#include "tea_text_document.h"
#include "interface.h"
#include "tea_gtk_utils.h"
#include "tea_config.h"
#include "tea_funx.h"
#include "rox_strings.h"


static gboolean pb_bounce (gpointer data)
{
  gtk_progress_bar_pulse (pb_progress_bar);
  return TRUE;
}


static void pb_start (void)
{
  id_pb = g_timeout_add (100, pb_progress_bar, NULL);
  gtk_widget_show (pb_progress_bar);
  gtk_progress_bar_pulse (pb_progress_bar);
}


static void pb_stop (void)
{
  gtk_progress_bar_set_fraction (pb_progress_bar, 0.0);
  g_source_remove (id_pb);
  gtk_widget_hide (pb_progress_bar);
}



static void on_button_encode (GtkWidget *wid, gpointer data)
{
  gchar *dir = gtk_entry_get_text (ent_dir_where_to_find);
  if (! dir)
      return;

  gchar *from = tv_get_selected_single (tv_charsets_from);
  gchar *to = tv_get_selected_single (tv_charsets_to);

  gboolean convert_eols = (gtk_toggle_button_get_active (&bt_convert_eol->toggle_button));
  gint eol_mode;

  if (gtk_toggle_button_get_active (bt_lf))
      eol_mode = eol_LF;
  if (gtk_toggle_button_get_active (bt_crlf))
      eol_mode = eol_CRLF;

  if (! from || ! to)
     return;

  gl_found_files = NULL;

  gsize bytes_read;
  gsize bytes_written;
  gchar *buf;
  gchar *dest;
  gchar *z;

  GError *error = NULL;
  GList *l;

  pb_start ();

  read_dir_files_recurse (dir, gtk_entry_get_text (ent_pattern));

  if (gl_found_files)
     {
      l = g_list_first (gl_found_files);
      while (l)
            {
             error = NULL;
             buf = str_file_read (l->data);
             if (buf) 
                {
                 dest = g_convert (buf, strlen (buf), to, from, &bytes_read, &bytes_written, &error);
                 if (dest)
                    {
                     if (convert_eols)
                        {
                         if (eol_mode == eol_LF) 
                            z = str_replace_all (dest, "\r\n", "\n");
                         if (eol_mode == eol_CRLF)
                            z = str_eol_to_crlf (dest);

                         save_string_to_file (l->data, z);

                         g_free (dest);
                         g_free (z);
                        }
                     else
                         {
                          save_string_to_file (l->data, dest);
                          g_free (dest);
                         }
                    }
                 g_free (buf);
                }
             l = g_list_next (l);
            }

     }

  pb_stop ();

  gtk_widget_show (ffstatusbar);
}



GtkWidget* create_wnd_encode (void)
{
  GtkWidget *window1 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window1), _("Encode files"));

  vbox1 = gtk_vbox_new (FALSE, UI_PACKFACTOR);
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (window1), vbox1);

  hbox1 = gtk_vbox_new (FALSE, UI_PACKFACTOR);
  gtk_widget_show (hbox1);
  gtk_box_pack_start (GTK_BOX (vbox1), hbox1, FALSE, FALSE, UI_PACKFACTOR);

  ent_dir_where_to_find = tea_dir_selector2 (hbox1, _("Directory"), g_get_home_dir ());

  ent_pattern = tea_text_entry (hbox1, _("File pattern"), "*.txt");


  GtkWidget *hbox5 = gtk_hbox_new (TRUE, UI_PACKFACTOR);
  gtk_widget_show (hbox5);
  gtk_box_pack_start (GTK_BOX (vbox1), hbox5, FALSE, FALSE, UI_PACKFACTOR);

  bt_convert_eol = gtk_check_button_new_with_label (_("Convert ends of lines"));
  gtk_widget_show (bt_convert_eol);
  gtk_box_pack_start (GTK_BOX (hbox5), bt_convert_eol, FALSE, FALSE, UI_PACKFACTOR);

  bt_lf = gtk_radio_button_new_with_label (NULL, "LF");
  bt_crlf = gtk_radio_button_new_with_label (gtk_radio_button_get_group (bt_lf), "CR/LF");

  gtk_widget_show (bt_lf);
  gtk_box_pack_start (GTK_BOX (hbox5), bt_lf, FALSE, FALSE, UI_PACKFACTOR);
  gtk_widget_show (bt_crlf);
  gtk_box_pack_start (GTK_BOX (hbox5), bt_crlf, FALSE, FALSE, UI_PACKFACTOR);


  GtkWidget *bt_encode = gtk_button_new_with_label (_("Find files matching the pattern and encode them")); 
  gtk_widget_show (bt_encode);
  gtk_box_pack_start (GTK_BOX (vbox1), bt_encode, FALSE, FALSE, UI_PACKFACTOR);

  g_signal_connect ((gpointer) bt_encode, "clicked",
                    G_CALLBACK (on_button_encode),
                    NULL);

  GtkWidget *hbox4 = gtk_hbox_new (TRUE, UI_PACKFACTOR);
  gtk_widget_show (hbox4);
  gtk_box_pack_start (GTK_BOX (vbox1), hbox4, TRUE, TRUE, UI_PACKFACTOR);

  tv_charsets_from = tv_create_framed (hbox4, _("From: "), GTK_SELECTION_SINGLE);
  tv_fill_with_glist (tv_charsets_from, gl_enc_available);

  tv_charsets_to = tv_create_framed (hbox4, _("To: "), GTK_SELECTION_SINGLE);
  tv_fill_with_glist (tv_charsets_to, gl_enc_available);



/*
  GtkWidget *bt_copy_to_dir = gtk_button_new_with_label (_("Copy to the directory")); 
  gtk_widget_show (bt_copy_to_dir);
  gtk_box_pack_start (GTK_BOX (hbox4), bt_copy_to_dir, FALSE, FALSE, UI_PACKFACTOR);

  g_signal_connect ((gpointer) bt_copy_to_dir, "clicked",
                    G_CALLBACK (on_button_copy_to_dir),
                    NULL);

  GtkWidget *bt_savesession = gtk_button_new_with_label (_("Save as a session")); 
  gtk_widget_show (bt_savesession);
  gtk_box_pack_start (GTK_BOX (hbox4), bt_savesession, FALSE, FALSE, UI_PACKFACTOR);

  g_signal_connect ((gpointer) bt_savesession, "clicked",
                    G_CALLBACK (on_button_savesession),
                    NULL);


  GtkWidget *hbox3 = gtk_hbox_new (TRUE, UI_PACKFACTOR);
  gtk_widget_show (hbox3);
  gtk_box_pack_start (GTK_BOX (vbox1), hbox3, FALSE, TRUE, UI_PACKFACTOR);


  tv_found_files = tv_create_framed (hbox1, _("Found files"), GTK_SELECTION_SINGLE);


  g_signal_connect (GTK_OBJECT (tv_found_files), "button_press_event",
                     (GtkSignalFunc) button_press_event, NULL);
*/

  ffstatusbar = gtk_statusbar_new ();
  gtk_widget_show (ffstatusbar);
  gtk_box_pack_start (GTK_BOX (vbox1), ffstatusbar, FALSE, FALSE, UI_PACKFACTOR);

  //b.p. Scorn - Whine - Strand

  pb_progress_bar = gtk_progress_bar_new ();
  gtk_box_pack_start (GTK_BOX (ffstatusbar), pb_progress_bar, FALSE, FALSE, UI_PACKFACTOR);
  gtk_progress_bar_set_pulse_step (pb_progress_bar, 0.05);
  gtk_widget_show (ffstatusbar);

//  gtk_window_set_focus (window1, ent_text_to_find);

  gtk_window_resize (window1, get_value (confile.screen_w, 80), get_value (confile.screen_h, 80));

//  gtk_window_set_modal (window1, TRUE);

  g_signal_connect (G_OBJECT (window1), "key_press_event", G_CALLBACK (win_key_handler), window1);

 g_signal_connect ((gpointer) window1, "delete_event",
                    G_CALLBACK (gtk_false),
                    NULL);

 // gtk_window_set_transient_for (window1, tea_main_window);
  gtk_window_set_position (window1, GTK_WIN_POS_CENTER_ON_PARENT);
  gtk_widget_show (window1);

  return window1;
}
